import os
import time
from concurrent import futures
from pathlib import Path
from typing import Any

import h5py
import numpy as np
import torch
from skimage import measure
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm

from pytorch3dunet.datasets.hdf5 import AbstractHDF5Dataset
from pytorch3dunet.datasets.utils import remove_padding
from pytorch3dunet.unet3d.config import TorchDevice
from pytorch3dunet.unet3d.model import is_model_2d
from pytorch3dunet.unet3d.utils import get_logger

logger = get_logger("UNetPredictor")


class AbstractPredictor:
    """Base class for predictors.

    Args:
        model: Segmentation model.
        output_dir: Directory where the predictions will be saved.
        out_channels: Number of output channels of the model.
        device: Device to run the model on (CPU, CUDA, MPS).
        output_dataset: Name of the dataset in the H5 file where the predictions will be saved. Default: 'predictions'.
        save_segmentation: If True the segmentation will be saved instead of the probability maps. Default: False.
        prediction_channel: Save only the specified channel from the network output. Default: None.
        performance_metric: Performance metric to be used for evaluation. Default: None.
        gt_internal_path: Internal path to the ground truth dataset in the H5 file. Default: None.
    """

    def __init__(
        self,
        model: nn.Module,
        output_dir: str,
        out_channels: int,
        device: TorchDevice,
        output_dataset: str = "predictions",
        save_segmentation: bool = False,
        prediction_channel: int | None = None,
        performance_metric: str | None = None,
        gt_internal_path: str | None = None,
        **kwargs,
    ):
        self.model = model
        self.output_dir = output_dir
        assert out_channels > 0, f"Invalid number of output channels: {out_channels}"
        self.out_channels = out_channels
        assert device in [TorchDevice.CPU, TorchDevice.CUDA, TorchDevice.MPS], f"Unsupported device: {device}"
        self.device = device
        self.output_dataset = output_dataset
        self.save_segmentation = save_segmentation
        self.prediction_channel = prediction_channel
        self.performance_metric = performance_metric
        self.gt_internal_path = gt_internal_path

    def __call__(self, test_loader: DataLoader) -> Any:
        """Run the model prediction on the test_loader and save the results in the output_dir.

        If the performance_metric is provided, the predictions will be evaluated against the ground truth
        and returned as a tensor.

        Args:
            test_loader: DataLoader for the test dataset.

        Returns:
            Evaluation metric result if performance_metric is provided, otherwise None.
        """
        raise NotImplementedError


class StandardPredictor(AbstractPredictor):
    """Standard predictor that applies the model on the given dataset and saves the result as H5 file.

    Predictions from the network are kept in memory. If the results from the network don't fit into RAM
    use `LazyPredictor` instead.
    """

    def __init__(
        self,
        model: nn.Module,
        output_dir: str,
        out_channels: int,
        device: TorchDevice,
        output_dataset: str = "predictions",
        save_segmentation: bool = False,
        prediction_channel: int | None = None,
        performance_metric: str | None = None,
        gt_internal_path: str | None = None,
        **kwargs,
    ):
        super().__init__(
            model,
            output_dir,
            out_channels,
            device,
            output_dataset,
            save_segmentation,
            prediction_channel,
            performance_metric,
            gt_internal_path,
            **kwargs,
        )

    def __call__(self, test_loader: DataLoader) -> Any:
        assert isinstance(test_loader.dataset, AbstractHDF5Dataset)
        logger.info(f"Processing '{test_loader.dataset.file_path}'...")
        start = time.perf_counter()

        # dimensionality of the output predictions
        volume_shape = test_loader.dataset.volume_shape

        if self.save_segmentation:
            # single channel segmentation map
            prediction_shape = volume_shape
        else:
            if self.prediction_channel is not None:
                # single channel prediction map
                prediction_shape = (1,) + volume_shape
            else:
                prediction_shape = (self.out_channels,) + volume_shape

        # create destination H5 file
        output_file = _get_output_file(dataset=test_loader.dataset, output_dir=self.output_dir)
        logger.info(f"Saving predictions to: {output_file}")
        with h5py.File(output_file, "w") as h5_output_file:
            # allocate prediction arrays
            logger.info("Allocating prediction arrays...")
            prediction_array = self._allocate_prediction_array(prediction_shape, h5_output_file)

            # determine halo used for padding
            patch_halo = test_loader.dataset.halo_shape
            logger.info(f"Using halo: {patch_halo}")

            # Sets the module in evaluation mode explicitly
            # It is necessary for batchnorm/dropout layers if present as well as final Sigmoid/Softmax to be applied
            self.model.eval()
            logger.info(f"Running inference on {len(test_loader)} batches")
            # Run predictions on the entire input dataset
            with torch.no_grad():
                for input, indices in tqdm(test_loader):
                    # send batch to device
                    # generally this is an effective choice for enhancing transfer speed,
                    # see: https://docs.pytorch.org/tutorials/intermediate/pinmem_nonblock.html
                    input = input.to(self.device, non_blocking=True)

                    if is_model_2d(self.model):
                        # remove the singleton z-dimension from the input
                        input = torch.squeeze(input, dim=-3)
                        # forward pass
                        prediction = self.model(input)
                        # add the singleton z-dimension to the output
                        prediction = torch.unsqueeze(prediction, dim=-3)
                    else:
                        # forward pass
                        prediction = self.model(input)

                    # unpad the predicted patch
                    if sum(patch_halo) > 0:
                        prediction = remove_padding(prediction, patch_halo)
                    # convert to numpy array
                    prediction = prediction.cpu().numpy()
                    # for each batch sample
                    for pred, index in zip(prediction, indices, strict=True):
                        if self.save_segmentation:
                            # if single channel, binarize
                            if pred.shape[0] == 1:
                                pred = pred[0] > 0.5
                            else:
                                # use the argmax of the prediction
                                pred = np.argmax(pred, axis=0)
                            pred = pred.astype("uint16")
                            index = tuple(index)
                        else:
                            # save patch index: (C,D,H,W)
                            if self.prediction_channel is None:
                                channel_slice = slice(0, self.out_channels)
                            else:
                                # use only the specified channel
                                channel_slice = slice(0, 1)
                                pred = np.expand_dims(pred[self.prediction_channel], axis=0)
                            # add channel dimension to the index
                            index = (channel_slice,) + tuple(index)

                        # accumulate probabilities into the output prediction array
                        prediction_array[index] = pred

            logger.info(f"Finished inference in {time.perf_counter() - start:.2f} seconds")
            # save results
            output_type = "segmentation" if self.save_segmentation else "probability maps"
            logger.info(f"Saving {output_type} to: {output_file}")
            self._create_prediction_dataset(h5_output_file, prediction_array)

            if self.performance_metric is not None:
                # load gt from the dataset
                assert self.gt_internal_path is not None
                gt = _load_dataset(test_loader.dataset, self.gt_internal_path)
                prediction_array = prediction_array[...]
                # create metric
                assert self.performance_metric in ["dice", "mean_iou"], (
                    f"Unsupported performance metric: {self.performance_metric}, only dice and mean_iou are supported"
                )
                if self.performance_metric == "dice":
                    result = dice_score(prediction_array, gt)
                else:
                    result = mean_iou(prediction_array, gt, n_classes=self.out_channels)
                return result

    def _create_prediction_dataset(self, h5_output_file, prediction_array):
        h5_output_file.create_dataset(self.output_dataset, data=prediction_array, compression="gzip")

    def _allocate_prediction_array(self, output_shape, output_file):
        if self.save_segmentation:
            dtype = "uint16"
        else:
            dtype = "float32"
        # initialize the output prediction arrays
        return np.zeros(output_shape, dtype=dtype)


class LazyPredictor(StandardPredictor):
    """Lazy predictor that applies the model on the given dataset and saves results in H5 format.

    Predicted patches are directly saved into the H5 file and won't be stored in memory. Since this predictor
    is slower than `StandardPredictor`, it should only be used when the predicted volume does not fit into RAM.
    """

    def __init__(
        self,
        model: nn.Module,
        output_dir: str,
        out_channels: int,
        device: TorchDevice,
        output_dataset: str = "predictions",
        save_segmentation: bool = False,
        prediction_channel: int | None = None,
        performance_metric: str | None = None,
        gt_internal_path: str | None = None,
        **kwargs,
    ):
        super().__init__(
            model,
            output_dir,
            out_channels,
            device,
            output_dataset,
            save_segmentation,
            prediction_channel,
            performance_metric,
            gt_internal_path,
            **kwargs,
        )

    def _allocate_prediction_array(self, output_shape, output_file):
        if self.save_segmentation:
            dtype = "uint16"
        else:
            dtype = "float32"
        # allocate datasets for probability maps
        prediction_array = output_file.create_dataset(
            self.output_dataset, shape=output_shape, dtype=dtype, chunks=True, compression="gzip"
        )
        return prediction_array

    def _create_prediction_dataset(self, h5_output_file, prediction_array):
        # no need to save the prediction array, it is already saved in the H5 file
        pass


class DSB2018Predictor(AbstractPredictor):
    def __init__(self, model, output_dir, config, save_segmentation=True, pmaps_thershold=0.5, **kwargs):
        super().__init__(model, output_dir, out_channels=1, device=config["device"], **kwargs)
        self.pmaps_threshold = pmaps_thershold
        self.save_segmentation = save_segmentation

    def _slice_from_pad(self, pad):
        if pad == 0:
            return slice(None, None)
        else:
            return slice(pad, -pad)

    def __call__(self, test_loader):
        # Sets the module in evaluation mode explicitly
        self.model.eval()
        # initial process pool for saving results to disk
        executor = futures.ProcessPoolExecutor(max_workers=32)
        # Run predictions on the entire input dataset
        with torch.no_grad():
            for img, path in test_loader:
                # send batch to gpu
                if self.device == TorchDevice.CUDA:
                    img = img.cuda(non_blocking=True)
                else:
                    img = img.to(self.device)
                # forward pass
                pred = self.model(img)

                executor.submit(dsb_save_batch, self.output_dir, path, pred)

        print("Waiting for all predictions to be saved to disk...")
        executor.shutdown(wait=True)


def dsb_save_batch(output_dir, path, pred, save_segmentation=True, pmaps_thershold=0.5):
    def _pmaps_to_seg(pred):
        mask = pred > pmaps_thershold
        return measure.label(mask).astype("uint16")

    # convert to numpy array
    for single_pred, single_path in zip(pred, path, strict=False):
        logger.info(f"Processing {single_path}")
        single_pred = single_pred.squeeze()

        # save to h5 file
        out_file = os.path.splitext(single_path)[0] + "_predictions.h5"
        if output_dir is not None:
            out_file = os.path.join(output_dir, os.path.split(out_file)[1])

        with h5py.File(out_file, "w") as f:
            # logger.info(f'Saving output to {out_file}')
            f.create_dataset("predictions", data=single_pred, compression="gzip")
            if save_segmentation:
                f.create_dataset("segmentation", data=_pmaps_to_seg(single_pred), compression="gzip")


def _get_output_file(
    dataset: AbstractHDF5Dataset, suffix: str = "_predictions", output_dir: str | Path | None = None
) -> Path:
    """
    Get the output file path for the predictions. If `output_dir` is not None the output file will be saved in
    the original dataset directory.

    Args:
        dataset: input dataset
        suffix: file name suffix
        output_dir: directory where the output file will be saved, if None the output file will be saved in the
            same directory as the input dataset

    Returns:
        path to the output file
    """
    file_path = Path(dataset.file_path)

    if output_dir is None:
        output_dir = file_path.parent
    else:
        output_dir = Path(output_dir)

    output_filename = file_path.stem + suffix + ".h5"
    return output_dir / output_filename


def _load_dataset(dataset: AbstractHDF5Dataset, internal_path: str) -> np.ndarray:
    file_path = dataset.file_path
    with h5py.File(file_path, "r") as f:
        return f[internal_path][...]


def mean_iou(pred: np.ndarray, gt: np.ndarray, n_classes: int, avg: bool = False) -> list[float] | float:
    """
    Compute the mean Intersection over Union (IoU) for the given predictions and ground truth.
    Args:
        pred: predicted segmentation
        gt: ground truth segmentation
        n_classes: number of classes
        avg: if True, return the mean IoU, otherwise return the IoU for each class
    Returns:
        mean IoU
    """
    # convert to numpy arrays
    pred = pred.astype("uint16")
    gt = gt.astype("uint16")
    assert pred.shape == gt.shape, f"Predictions and ground truth have different shapes: {pred.shape} != {gt.shape}"

    # compute IoU, skip 0: background
    per_class_iou = []
    for c in range(1, n_classes):
        intersection = np.logical_and(gt == c, pred == c).sum()
        union = np.logical_or(gt == c, pred == c).sum()
        iou = intersection / union
        per_class_iou.append(iou)

    if avg:
        return np.mean(per_class_iou)
    return per_class_iou


def dice_score(pred: np.ndarray, gt: np.ndarray, avg: bool = False) -> list[float] | float:
    """
    Compute the Dice score for the given predictions and ground truth.
    If avg is True, return the mean Dice score, otherwise return the Dice score for each channel/class.
    """
    # convert to numpy arrays
    pred = pred.astype("uint16")
    gt = gt.astype("uint16")
    assert pred.shape == gt.shape, f"Predictions and ground truth have different shapes: {pred.shape} != {gt.shape}"
    # compute Dice score
    per_class_dice = []
    for c_pred, c_gt in zip(pred, gt, strict=True):
        intersection = np.logical_and(c_gt, c_pred).sum()
        union = c_gt.sum() + c_pred.sum()
        dice = 2 * intersection / union
        per_class_dice.append(dice)
    if avg:
        return np.mean(per_class_dice)
    return per_class_dice



from functools import partial

import torch
from torch import nn as nn
from torch.nn import functional as F

from pytorch3dunet.unet3d.se import ChannelSELayer3D, ChannelSpatialSELayer3D, SpatialSELayer3D


def create_conv(
    in_channels: int,
    out_channels: int,
    kernel_size: int | tuple[int],
    order: str,
    num_groups: int,
    padding: int | tuple[int],
    dropout_prob: float,
    is3d: bool,
) -> list[tuple[str, nn.Module]]:
    """
    Create a list of modules for a given level of UNet network. It consists of a single conv layer with non-linearity
    and optional batchnorm/groupnorm.

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size(int or tuple): size of the convolving kernel
        order (str): order of things, e.g.
            'cr' -> conv + ReLU
            'gcr' -> groupnorm + conv + ReLU
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
            'bcr' -> batchnorm + conv + ReLU
        num_groups (int): number of groups for the GroupNorm
        padding (int or tuple): add zero-padding added to all three sides of the input
        dropout_prob (float): dropout probability
        is3d (bool): is3d (bool): if True use Conv3d, otherwise use Conv2d
    Return:
        list modules, where each module is a tuple (name, module)
    """
    assert "c" in order, "Conv layer MUST be present"
    assert order[0] not in "rle", "Non-linearity cannot be the first operation in the layer"

    modules = []
    for i, char in enumerate(order):
        if char == "r":
            modules.append(("ReLU", nn.ReLU(inplace=True)))
        elif char == "l":
            modules.append(("LeakyReLU", nn.LeakyReLU(inplace=True)))
        elif char == "e":
            modules.append(("ELU", nn.ELU(inplace=True)))
        elif char == "c":
            # add learnable bias only in the absence of batchnorm/groupnorm
            bias = not ("g" in order or "b" in order)
            if is3d:
                conv = nn.Conv3d(in_channels, out_channels, kernel_size, padding=padding, bias=bias)
            else:
                conv = nn.Conv2d(in_channels, out_channels, kernel_size, padding=padding, bias=bias)

            modules.append(("conv", conv))
        elif char == "g":
            is_before_conv = i < order.index("c")
            if is_before_conv:
                num_channels = in_channels
            else:
                num_channels = out_channels

            # use only one group if the given number of groups is greater than the number of channels
            if num_channels < num_groups:
                num_groups = 1

            assert num_channels % num_groups == 0, (
                f"Expected number of channels in input to be divisible by num_groups. num_channels={num_channels}, num_groups={num_groups}"
            )
            modules.append(("groupnorm", nn.GroupNorm(num_groups=num_groups, num_channels=num_channels)))
        elif char == "b":
            is_before_conv = i < order.index("c")
            if is3d:
                bn = nn.BatchNorm3d
            else:
                bn = nn.BatchNorm2d

            if is_before_conv:
                modules.append(("batchnorm", bn(in_channels)))
            else:
                modules.append(("batchnorm", bn(out_channels)))
        elif char == "d":
            modules.append(("dropout", nn.Dropout(p=dropout_prob)))
        elif char == "D":
            modules.append(("dropout2d", nn.Dropout2d(p=dropout_prob)))
        else:
            raise ValueError(
                f"Unsupported layer type '{char}'. MUST be one of ['b', 'g', 'r', 'l', 'e', 'c', 'd', 'D']"
            )

    return modules


class SingleConv(nn.Sequential):
    """
    Basic convolutional module consisting of a Conv3d, non-linearity and optional batchnorm/groupnorm. The order
    of operations can be specified via the `order` parameter

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size (int or tuple): size of the convolving kernel
        order (string): determines the order of layers, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
        num_groups (int): number of groups for the GroupNorm
        padding (int or tuple): add zero-padding
        dropout_prob (float): dropout probability, default 0.1
        is3d (bool): if True use Conv3d, otherwise use Conv2d
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        order="gcr",
        num_groups=8,
        padding=1,
        dropout_prob=0.1,
        is3d=True,
    ):
        super().__init__()

        for name, module in create_conv(
            in_channels, out_channels, kernel_size, order, num_groups, padding, dropout_prob, is3d
        ):
            self.add_module(name, module)


class DoubleConv(nn.Sequential):
    """
    A module consisting of two consecutive convolution layers. We use 2x (Conv3d+ReLU+GroupNorm3d) by default.
    This can be changed however by providing the 'order' argument, e.g. in order
    to change to Conv3d+BatchNorm3d+ELU use order='cbe'.
    Use padded convolutions to make sure that the output (H_out, W_out) is the same
    as (H_in, W_in), so that you don't have to crop in the decoder path.

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        encoder (bool): if True we're in the encoder path, otherwise we're in the decoder
        kernel_size (int or tuple): size of the convolving kernel
        order (string): determines the order of layers, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
        num_groups (int): number of groups for the GroupNorm
        padding (int or tuple): add zero-padding added to all three sides of the input
        upscale (int): number of the convolution to upscale in encoder if DoubleConv, default: 2
        dropout_prob (float or tuple): dropout probability for each convolution, default 0.1
        is3d (bool): if True use Conv3d instead of Conv2d layers
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        encoder,
        kernel_size=3,
        order="gcr",
        num_groups=8,
        padding=1,
        upscale=2,
        dropout_prob=0.1,
        is3d=True,
    ):
        super().__init__()
        if encoder:
            # we're in the encoder path
            conv1_in_channels = in_channels
            if upscale == 1:
                conv1_out_channels = out_channels
            else:
                conv1_out_channels = out_channels // 2
            if conv1_out_channels < in_channels:
                conv1_out_channels = in_channels
            conv2_in_channels, conv2_out_channels = conv1_out_channels, out_channels
        else:
            # we're in the decoder path, decrease the number of channels in the 1st convolution
            conv1_in_channels, conv1_out_channels = in_channels, out_channels
            conv2_in_channels, conv2_out_channels = out_channels, out_channels

        # check if dropout_prob is a tuple and if so
        # split it for different dropout probabilities for each convolution.
        if isinstance(dropout_prob, list) or isinstance(dropout_prob, tuple):
            dropout_prob1 = dropout_prob[0]
            dropout_prob2 = dropout_prob[1]
        else:
            dropout_prob1 = dropout_prob2 = dropout_prob

        # conv1
        self.add_module(
            "SingleConv1",
            SingleConv(
                conv1_in_channels,
                conv1_out_channels,
                kernel_size,
                order,
                num_groups,
                padding=padding,
                dropout_prob=dropout_prob1,
                is3d=is3d,
            ),
        )
        # conv2
        self.add_module(
            "SingleConv2",
            SingleConv(
                conv2_in_channels,
                conv2_out_channels,
                kernel_size,
                order,
                num_groups,
                padding=padding,
                dropout_prob=dropout_prob2,
                is3d=is3d,
            ),
        )


class ResNetBlock(nn.Module):
    """Residual block that can be used instead of standard DoubleConv in the Encoder module.

    Motivated by: https://arxiv.org/pdf/1706.00120.pdf
    Notice we use ELU instead of ReLU (order='cge') and put non-linearity after the groupnorm.

    Args:
        in_channels: Number of input channels.
        out_channels: Number of output channels.
        kernel_size: Size of the convolving kernel. Default: 3.
        order: Determines the order of layers. Default: 'cge'.
        num_groups: Number of groups for the GroupNorm. Default: 8.
        is3d: If True use Conv3d, otherwise use Conv2d. Default: True.
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, order="cge", num_groups=8, is3d=True, **kwargs):
        super().__init__()

        if in_channels != out_channels:
            # conv1x1 for increasing the number of channels
            if is3d:
                self.conv1 = nn.Conv3d(in_channels, out_channels, 1)
            else:
                self.conv1 = nn.Conv2d(in_channels, out_channels, 1)
        else:
            self.conv1 = nn.Identity()

        # residual block
        self.conv2 = SingleConv(
            out_channels, out_channels, kernel_size=kernel_size, order=order, num_groups=num_groups, is3d=is3d
        )
        # remove non-linearity from the 3rd convolution since it's going to be applied after adding the residual
        n_order = order
        for c in "rel":
            n_order = n_order.replace(c, "")
        self.conv3 = SingleConv(
            out_channels, out_channels, kernel_size=kernel_size, order=n_order, num_groups=num_groups, is3d=is3d
        )

        # create non-linearity separately
        if "l" in order:
            self.non_linearity = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        elif "e" in order:
            self.non_linearity = nn.ELU(inplace=True)
        else:
            self.non_linearity = nn.ReLU(inplace=True)

    def forward(self, x):
        # apply first convolution to bring the number of channels to out_channels
        residual = self.conv1(x)

        # residual block
        out = self.conv2(residual)
        out = self.conv3(out)

        out += residual
        out = self.non_linearity(out)

        return out


class ResNetBlockSE(ResNetBlock):
    def __init__(self, in_channels, out_channels, kernel_size=3, order="cge", num_groups=8, se_module="scse", **kwargs):
        super().__init__(
            in_channels, out_channels, kernel_size=kernel_size, order=order, num_groups=num_groups, **kwargs
        )
        assert se_module in ["scse", "cse", "sse"]
        if se_module == "scse":
            self.se_module = ChannelSpatialSELayer3D(num_channels=out_channels, reduction_ratio=1)
        elif se_module == "cse":
            self.se_module = ChannelSELayer3D(num_channels=out_channels, reduction_ratio=1)
        elif se_module == "sse":
            self.se_module = SpatialSELayer3D(num_channels=out_channels)

    def forward(self, x):
        out = super().forward(x)
        out = self.se_module(out)
        return out


class Encoder(nn.Module):
    """
    A single module from the encoder path consisting of the optional max
    pooling layer (one may specify the MaxPool kernel_size to be different
    from the standard (2,2,2), e.g. if the volumetric data is anisotropic
    (make sure to use complementary scale_factor in the decoder path) followed by
    a basic module (DoubleConv or ResNetBlock).

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        conv_kernel_size (int or tuple): size of the convolving kernel
        apply_pooling (bool): if True use MaxPool3d before DoubleConv
        pool_kernel_size (int or tuple): the size of the window
        pool_type (str): pooling layer: 'max' or 'avg'
        basic_module(nn.Module): either ResNetBlock or DoubleConv
        conv_layer_order (string): determines the order of layers
            in `DoubleConv` module. See `DoubleConv` for more info.
        num_groups (int): number of groups for the GroupNorm
        padding (int or tuple): add zero-padding added to all three sides of the input
        upscale (int): number of the convolution to upscale in encoder if DoubleConv, default: 2
        dropout_prob (float or tuple): dropout probability, default 0.1
        is3d (bool): use 3d or 2d convolutions/pooling operation
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        conv_kernel_size=3,
        apply_pooling=True,
        pool_kernel_size=2,
        pool_type="max",
        basic_module=DoubleConv,
        conv_layer_order="gcr",
        num_groups=8,
        padding=1,
        upscale=2,
        dropout_prob=0.1,
        is3d=True,
    ):
        super().__init__()
        assert pool_type in ["max", "avg"]
        if apply_pooling:
            if pool_type == "max":
                if is3d:
                    self.pooling = nn.MaxPool3d(kernel_size=pool_kernel_size)
                else:
                    self.pooling = nn.MaxPool2d(kernel_size=pool_kernel_size)
            else:
                if is3d:
                    self.pooling = nn.AvgPool3d(kernel_size=pool_kernel_size)
                else:
                    self.pooling = nn.AvgPool2d(kernel_size=pool_kernel_size)
        else:
            self.pooling = None

        self.basic_module = basic_module(
            in_channels,
            out_channels,
            encoder=True,
            kernel_size=conv_kernel_size,
            order=conv_layer_order,
            num_groups=num_groups,
            padding=padding,
            upscale=upscale,
            dropout_prob=dropout_prob,
            is3d=is3d,
        )

    def forward(self, x):
        if self.pooling is not None:
            x = self.pooling(x)
        x = self.basic_module(x)
        return x


class Decoder(nn.Module):
    """
    A single module for decoder path consisting of the upsampling layer
    (either learned ConvTranspose3d or nearest neighbor interpolation)
    followed by a basic module (DoubleConv or ResNetBlock).

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        conv_kernel_size (int or tuple): size of the convolving kernel
        scale_factor (int or tuple): used as the multiplier for the image H/W/D in
            case of nn.Upsample or as stride in case of ConvTranspose3d, must reverse the MaxPool3d operation
            from the corresponding encoder
        basic_module(nn.Module): either ResNetBlock or DoubleConv
        conv_layer_order (string): determines the order of layers
            in `DoubleConv` module. See `DoubleConv` for more info.
        num_groups (int): number of groups for the GroupNorm
        padding (int or tuple): add zero-padding added to all three sides of the input
        upsample (str): algorithm used for upsampling:
            InterpolateUpsampling:   'nearest' | 'linear' | 'bilinear' | 'trilinear' | 'area'
            TransposeConvUpsampling: 'deconv'
            No upsampling:           None
            Default: 'default' (chooses automatically)
        dropout_prob (float or tuple): dropout probability, default 0.1
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        conv_kernel_size=3,
        scale_factor=2,
        basic_module=DoubleConv,
        conv_layer_order="gcr",
        num_groups=8,
        padding=1,
        upsample="default",
        dropout_prob=0.1,
        is3d=True,
    ):
        super().__init__()

        # perform concat joining per default
        concat = True

        # don't adapt channels after join operation
        adapt_channels = False

        if upsample is not None and upsample != "none":
            if upsample == "default":
                if basic_module == DoubleConv:
                    upsample = "nearest"  # use nearest neighbor interpolation for upsampling
                    concat = True  # use concat joining
                    adapt_channels = False  # don't adapt channels
                elif basic_module == ResNetBlock or basic_module == ResNetBlockSE:
                    upsample = "deconv"  # use deconvolution upsampling
                    concat = False  # use summation joining
                    adapt_channels = True  # adapt channels after joining

            # perform deconvolution upsampling if mode is deconv
            if upsample == "deconv":
                self.upsampling = TransposeConvUpsampling(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=conv_kernel_size,
                    scale_factor=scale_factor,
                    is3d=is3d,
                )
            else:
                self.upsampling = InterpolateUpsampling(mode=upsample)
        else:
            # no upsampling
            self.upsampling = NoUpsampling()
            # concat joining
            self.joining = partial(self._joining, concat=True)

        # perform joining operation
        self.joining = partial(self._joining, concat=concat)

        # adapt the number of in_channels for the ResNetBlock
        if adapt_channels is True:
            in_channels = out_channels

        self.basic_module = basic_module(
            in_channels,
            out_channels,
            encoder=False,
            kernel_size=conv_kernel_size,
            order=conv_layer_order,
            num_groups=num_groups,
            padding=padding,
            dropout_prob=dropout_prob,
            is3d=is3d,
        )

    def forward(self, encoder_features, x):
        x = self.upsampling(encoder_features=encoder_features, x=x)
        x = self.joining(encoder_features, x)
        x = self.basic_module(x)
        return x

    @staticmethod
    def _joining(encoder_features, x, concat):
        if concat:
            return torch.cat((encoder_features, x), dim=1)
        else:
            return encoder_features + x


def create_encoders(
    in_channels,
    f_maps,
    basic_module,
    conv_kernel_size,
    conv_padding,
    conv_upscale,
    dropout_prob,
    layer_order,
    num_groups,
    pool_kernel_size,
    is3d,
):
    # create encoder path consisting of Encoder modules. Depth of the encoder is equal to `len(f_maps)`
    encoders = []
    for i, out_feature_num in enumerate(f_maps):
        if i == 0:
            # apply conv_coord only in the first encoder if any
            encoder = Encoder(
                in_channels,
                out_feature_num,
                apply_pooling=False,  # skip pooling in the firs encoder
                basic_module=basic_module,
                conv_layer_order=layer_order,
                conv_kernel_size=conv_kernel_size,
                num_groups=num_groups,
                padding=conv_padding,
                upscale=conv_upscale,
                dropout_prob=dropout_prob,
                is3d=is3d,
            )
        else:
            encoder = Encoder(
                f_maps[i - 1],
                out_feature_num,
                basic_module=basic_module,
                conv_layer_order=layer_order,
                conv_kernel_size=conv_kernel_size,
                num_groups=num_groups,
                pool_kernel_size=pool_kernel_size,
                padding=conv_padding,
                upscale=conv_upscale,
                dropout_prob=dropout_prob,
                is3d=is3d,
            )

        encoders.append(encoder)

    return nn.ModuleList(encoders)


def create_decoders(
    f_maps, basic_module, conv_kernel_size, conv_padding, layer_order, num_groups, upsample, dropout_prob, is3d
):
    # create decoder path consisting of the Decoder modules. The length of the decoder list is equal to `len(f_maps) - 1`
    decoders = []
    reversed_f_maps = list(reversed(f_maps))
    for i in range(len(reversed_f_maps) - 1):
        if basic_module == DoubleConv and upsample != "deconv":
            in_feature_num = reversed_f_maps[i] + reversed_f_maps[i + 1]
        else:
            in_feature_num = reversed_f_maps[i]

        out_feature_num = reversed_f_maps[i + 1]

        decoder = Decoder(
            in_feature_num,
            out_feature_num,
            basic_module=basic_module,
            conv_layer_order=layer_order,
            conv_kernel_size=conv_kernel_size,
            num_groups=num_groups,
            padding=conv_padding,
            upsample=upsample,
            dropout_prob=dropout_prob,
            is3d=is3d,
        )
        decoders.append(decoder)
    return nn.ModuleList(decoders)


class AbstractUpsampling(nn.Module):
    """Abstract class for upsampling.

    A given implementation should upsample a given 5D input tensor using either
    interpolation or learned transposed convolution.

    Args:
        upsample: Upsampling function to be used.
    """

    def __init__(self, upsample):
        super().__init__()
        self.upsample = upsample

    def forward(self, encoder_features, x):
        # get the spatial dimensions of the output given the encoder_features
        output_size = encoder_features.size()[2:]
        # upsample the input and return
        return self.upsample(x, output_size)


class InterpolateUpsampling(AbstractUpsampling):
    """
    Non-learnable upsampling backed by interpolation.

    Args:
        mode (str): algorithm used for upsampling:
            'nearest' | 'linear' | 'bilinear' | 'trilinear' | 'area'. Default: 'nearest'
            used only if transposed_conv is False
    """

    def __init__(self, mode="nearest"):
        upsample = partial(self._interpolate, mode=mode)
        super().__init__(upsample)

    @staticmethod
    def _interpolate(x, size, mode):
        return F.interpolate(x, size=size, mode=mode)


class TransposeConvUpsampling(AbstractUpsampling):
    """
    Learned upsampling backed by transposed convolution layer followed by interpolation to the correct size if necessary.

    Args:
        in_channels (int): number of input channels for transposed conv
            used only if transposed_conv is True
        out_channels (int): number of output channels for transpose conv
            used only if transposed_conv is True
        kernel_size (int or tuple): size of the convolving kernel
            used only if transposed_conv is True
        scale_factor (int or tuple): stride of the convolution
            used only if transposed_conv is True
        is3d (bool): if True use ConvTranspose3d, otherwise use ConvTranspose2d
    """

    class Upsample(nn.Module):
        """Workaround for ValueError in transposed convolution.

        Performs transposed conv followed by interpolation to the correct size if necessary.
        This addresses the 'ValueError: requested an output size...' in the `_output_padding` method.

        Args:
            conv_transposed: Transposed convolution layer.
            is3d: If True use 3D operations, otherwise use 2D.
        """

        def __init__(self, conv_transposed, is3d):
            super().__init__()
            self.conv_transposed = conv_transposed
            self.is3d = is3d

        def forward(self, x, size):
            x = self.conv_transposed(x)
            return F.interpolate(x, size=size)

    def __init__(self, in_channels, out_channels, kernel_size=3, scale_factor=2, is3d=True):
        # make sure that the output size reverses the MaxPool3d from the corresponding encoder
        if is3d is True:
            conv_transposed = nn.ConvTranspose3d(
                in_channels, out_channels, kernel_size=kernel_size, stride=scale_factor, padding=1, bias=False
            )
        else:
            conv_transposed = nn.ConvTranspose2d(
                in_channels, out_channels, kernel_size=kernel_size, stride=scale_factor, padding=1, bias=False
            )
        upsample = self.Upsample(conv_transposed, is3d)
        super().__init__(upsample)


class NoUpsampling(AbstractUpsampling):
    """No upsampling, return the input as is."""

    def __init__(self):
        super().__init__(self._no_upsampling)

    @staticmethod
    def _no_upsampling(x, size):
        return x



import os
from pathlib import Path
from tempfile import NamedTemporaryFile

import h5py
import numpy as np
from torch.utils.data import DataLoader

from pytorch3dunet.datasets.hdf5 import LazyHDF5Dataset, StandardHDF5Dataset, traverse_h5_paths
from pytorch3dunet.datasets.utils import default_prediction_collate, remove_padding


class TestHDF5Dataset:
    def test_hdf5_dataset(self, transformer_config):
        path = create_random_dataset((128, 128, 128))

        patch_shapes = [(127, 127, 127), (69, 70, 70), (32, 64, 64)]
        stride_shapes = [(127, 127, 127), (69, 69, 69), (32, 64, 64)]

        phase = "test"

        for patch_shape, stride_shape in zip(patch_shapes, stride_shapes, strict=True):
            with h5py.File(path, "r") as f:
                raw = f["raw"][...]
                label = f["label"][...]

                dataset = StandardHDF5Dataset(
                    path,
                    phase=phase,
                    slice_builder_config=_slice_builder_conf(patch_shape, stride_shape),
                    transformer_config=transformer_config[phase]["transformer"],
                    raw_internal_path="raw",
                    label_internal_path="label",
                )

                # create zero-arrays of the same shape as the original dataset in order to verify if every element
                # was visited during the iteration
                visit_raw = np.zeros_like(raw)
                visit_label = np.zeros_like(label)

                for _, idx in dataset:
                    visit_raw[idx] = 1
                    visit_label[idx] = 1

                # verify that every element was visited at least once
                assert np.all(visit_raw)
                assert np.all(visit_label)

    def test_lazy_hdf5_dataset(self, transformer_config):
        path = create_random_dataset((128, 128, 128))

        patch_shapes = [(127, 127, 127), (69, 70, 70), (32, 64, 64)]
        stride_shapes = [(127, 127, 127), (69, 69, 69), (32, 64, 64)]
        halo_shape = (16, 32, 32)
        phase = "test"

        for patch_shape, stride_shape in zip(patch_shapes, stride_shapes, strict=True):
            with h5py.File(path, "r") as f:
                raw_shape = f["raw"].shape
                label_shape = f["label"].shape

            dataset = LazyHDF5Dataset(
                path,
                phase=phase,
                slice_builder_config=_slice_builder_conf(patch_shape, stride_shape, halo_shape),
                transformer_config=transformer_config[phase]["transformer"],
                raw_internal_path="raw",
                label_internal_path="label",
            )

            # create zero-arrays of the same shape as the original dataset in order to verify if every element
            # was visited during the iteration
            visit_raw = np.zeros(raw_shape)
            visit_label = np.zeros(label_shape)

            for _, idx in dataset:
                visit_raw[idx] = 1
                visit_label[idx] = 1

            # verify that every element was visited at least once
            assert np.all(visit_raw)
            assert np.all(visit_label)

    def test_augmentation(self, transformer_config, tmpdir):
        raw = np.random.rand(32, 96, 96)
        # assign raw to label's channels for ease of comparison
        label = np.stack([raw for _ in range(3)])
        # create temporary h5 file
        tmp_file = tmpdir / "test.h5"
        with h5py.File(tmp_file, "w") as f:
            f.create_dataset("raw", data=raw)
            f.create_dataset("label", data=label)

        # set phase='train' in order to execute the train transformers
        phase = "train"
        dataset = StandardHDF5Dataset(
            tmp_file,
            phase=phase,
            slice_builder_config=_slice_builder_conf((16, 64, 64), (8, 32, 32)),
            transformer_config=transformer_config[phase]["transformer"],
        )

        data_loader = DataLoader(dataset, batch_size=1, shuffle=True)
        for img, label in data_loader:
            for i in range(label.shape[0]):
                assert np.allclose(img, label[i])

    def test_traverse_file_paths(self, tmpdir):
        test_tmp_dir = os.path.join(tmpdir, "test")
        os.mkdir(test_tmp_dir)

        expected_files = [
            os.path.join(tmpdir, "f1.h5"),
            os.path.join(test_tmp_dir, "f2.h5"),
            os.path.join(test_tmp_dir, "f3.hdf"),
            os.path.join(test_tmp_dir, "f4.hdf5"),
            os.path.join(test_tmp_dir, "f5.hd5"),
        ]
        # create expected files
        for ef in expected_files:
            with h5py.File(ef, "w") as f:
                f.create_dataset("raw", data=np.random.randn(4, 4, 4))

        # make sure that traverse_file_paths runs correctly
        file_paths = [os.path.join(tmpdir, "f1.h5"), test_tmp_dir]
        actual_files = traverse_h5_paths(file_paths)

        assert expected_files == actual_files

    def test_halo(self, tmpdir: Path):
        halo_shape = (1, 2, 3)

        # create temporary h5 file
        raw = np.random.rand(32, 96, 96)
        tmp_file = tmpdir / "test.h5"

        with h5py.File(tmp_file, "w") as f:
            f.create_dataset("raw", data=raw)

        # halo only implemented with test phase
        phase = "test"
        dataset = StandardHDF5Dataset(
            tmp_file,
            phase=phase,
            slice_builder_config=_slice_builder_conf((16, 64, 64), (8, 32, 32), halo_shape),
            transformer_config=_transformer_test_conf(),
        )
        data_loader = DataLoader(dataset, batch_size=1, shuffle=True, collate_fn=default_prediction_collate)

        # verify all patches have the correct halo added and removed
        for input_batch, indices_batch in data_loader:  # input_batch has NCDHW shape, indices_batch has length N
            for input_, indices in zip(
                input_batch, indices_batch, strict=True
            ):  # input_ has CDHW shape, indices is for DHW
                input_ = remove_padding(input_, halo_shape)
                assert np.allclose(input_[0], raw[indices])

    def test_random_scale(self, transformer_config):
        path = create_random_dataset((200, 200, 172))

        patch_shapes = [(172, 172, 172)]
        stride_shapes = [(28, 28, 28)]

        phase = "train"

        for patch_shape, stride_shape in zip(patch_shapes, stride_shapes, strict=True):
            dataset = StandardHDF5Dataset(
                path,
                phase=phase,
                slice_builder_config=_slice_builder_conf(patch_shape, stride_shape),
                transformer_config=transformer_config[phase]["transformer"],
                raw_internal_path="raw",
                label_internal_path="label",
                random_scale=20,
            )

            for raw, label in dataset:
                if raw.ndim == 3:
                    assert raw.shape == patch_shape
                    assert label.shape == patch_shape
                else:
                    assert raw.shape[1:] == patch_shape
                    assert label.shape[1:] == patch_shape


def create_random_dataset(shape, ignore_index=False, raw_datasets=None, label_datasets=None):
    if label_datasets is None:
        label_datasets = ["label"]
    if raw_datasets is None:
        raw_datasets = ["raw"]

    tmp_file = NamedTemporaryFile(delete=False)

    with h5py.File(tmp_file.name, "w") as f:
        for raw_dataset in raw_datasets:
            f.create_dataset(raw_dataset, data=np.random.rand(*shape))

        for label_dataset in label_datasets:
            if ignore_index:
                f.create_dataset(label_dataset, data=np.random.randint(-1, 2, shape))
            else:
                f.create_dataset(label_dataset, data=np.random.randint(0, 2, shape))

    return tmp_file.name


def _slice_builder_conf(patch_shape, stride_shape, halo_shape=(0, 0, 0)):
    return {
        "name": "SliceBuilder",
        "patch_shape": patch_shape,
        "stride_shape": stride_shape,
        "halo_shape": halo_shape,
    }


def _transformer_test_conf():
    return {
        "raw": [{"name": "ToTensor", "expand_dims": True}],
        # 'label': [{'name': 'ToTensor', 'expand_dims': True}],
    }



import torch

from pytorch3dunet.unet3d.buildingblocks import ResNetBlock
from pytorch3dunet.unet3d.model import ResidualUNet2D, ResidualUNet3D, ResidualUNetSE3D, UNet2D, UNet3D


class TestModel:
    def test_unet2d(self):
        model = UNet2D(1, 1, f_maps=16, final_sigmoid=True)
        model.eval()
        x = torch.rand(1, 1, 65, 65)
        with torch.no_grad():
            y = model(x)

        assert torch.all(0 <= y) and torch.all(y <= 1)

    def test_unet3d(self):
        model = UNet3D(1, 1, f_maps=16, final_sigmoid=True)
        model.eval()
        x = torch.rand(1, 1, 33, 65, 65)
        with torch.no_grad():
            y = model(x)

        assert torch.all(0 <= y) and torch.all(y <= 1)

    def test_resnet_block1(self):
        blk = ResNetBlock(33, 64, is3d=False, order="cgr")
        blk.eval()
        x = torch.rand(1, 33, 65, 65)

        with torch.no_grad():
            y = blk(x)

        assert torch.all(0 <= y)

    def test_resnet_block2(self):
        blk = ResNetBlock(33, 32, is3d=False, order="cgr")
        blk.eval()
        x = torch.rand(1, 33, 65, 65)

        with torch.no_grad():
            y = blk(x)

        assert torch.all(0 <= y)

    def test_resunet3d(self):
        model = ResidualUNet3D(1, 1, f_maps=16, final_sigmoid=True)
        model.eval()
        x = torch.rand(1, 1, 33, 65, 65)
        y = model(x)

        assert torch.all(0 <= y) and torch.all(y <= 1)

    def test_resunet2d(self):
        model = ResidualUNet2D(1, 1, f_maps=16, final_sigmoid=True)
        model.eval()
        x = torch.rand(1, 1, 65, 65)
        with torch.no_grad():
            y = model(x)

        assert torch.all(0 <= y) and torch.all(y <= 1)

    def test_resunetSE3d(self):
        model = ResidualUNetSE3D(1, 1, f_maps=16, final_sigmoid=True)
        model.eval()
        x = torch.rand(1, 1, 33, 65, 65)
        y = model(x)

        assert torch.all(0 <= y) and torch.all(y <= 1)



import collections
from typing import Any

import h5py
import numpy as np
import torch
from torch.nn.functional import interpolate
from torch.utils.data import ConcatDataset, DataLoader, Dataset

from pytorch3dunet.unet3d.config import TorchDevice, os_dependent_dataloader_kwargs
from pytorch3dunet.unet3d.utils import get_class, get_logger

logger = get_logger("Dataset")


class RandomScaler:
    """
    Randomly scales the raw and label patches.

    Args:
        scale_range (int): the maximum absolute value of the scaling factor,
            i.e. patches coordinates will be randomly shifted in the range [-scale_range, scale_range]
        patch_shape (tuple): the shape of the patch DxHxW
        volume_shape (tuple): the shape of the volume DxHxW
        execution_probability (float): the probability of executing the scaling
        seed (int): random seed
    """

    def __init__(
        self,
        scale_range: int,
        patch_shape: tuple,
        volume_shape: tuple,
        execution_probability: float = 0.5,
        seed: int = 47,
    ):
        self.scale_range = scale_range
        self.patch_shape = patch_shape
        self.volume_shape = volume_shape
        self.execution_probability = execution_probability
        self.rs = np.random.RandomState(seed)

    def randomize_indices(self, raw_idx: tuple, label_idx: tuple) -> tuple[tuple, tuple]:
        # execute scaling with a given probability
        if self.rs.uniform() < self.execution_probability:
            return raw_idx, label_idx

        # select random offsets for scaling
        offsets = [self.rs.randint(self.scale_range) for _ in range(3)]
        # change offset sign at random
        if self.rs.rand() > 0.5:
            offsets = [-o for o in offsets]
        # apply offsets to the start or end of the slice at random
        is_start = self.rs.rand() > 0.5
        raw_idx = self._apply_offsets(raw_idx, offsets, is_start)
        label_idx = self._apply_offsets(label_idx, offsets, is_start)

        # assert spatial dimensions are the same
        if len(raw_idx) == 4:
            raw_idx_spacial = raw_idx[1:]
        else:
            raw_idx_spacial = raw_idx
        if len(label_idx) == 4:
            label_idx_spacial = label_idx[1:]
        else:
            label_idx_spacial = label_idx
        assert raw_idx_spacial == label_idx_spacial, (
            f"Raw and label indices are different: {raw_idx_spacial} != {label_idx_spacial}"
        )

        return raw_idx, label_idx

    def rescale_patches(self, raw_patch: torch.Tensor, label_patch: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        # compute zoom factors
        if raw_patch.ndim == 4:
            raw_shape = raw_patch.shape[1:]
        else:
            raw_shape = raw_patch.shape

        # if raw_shape equal to self.patch_shape just return the patches
        if raw_shape == self.patch_shape:
            return raw_patch, label_patch

        # rescale patches back to the original shape
        if raw_patch.ndim == 4:
            # add batch dimension
            raw_patch = raw_patch.unsqueeze(0)
            remove_dims = 1
        else:
            # add batch and channels dimensions
            raw_patch = raw_patch.unsqueeze(0).unsqueeze(0)
            remove_dims = 2

        # interpolate raw patch
        raw_patch = interpolate(raw_patch, self.patch_shape, mode="trilinear")
        # remove additional dimensions
        for _ in range(remove_dims):
            raw_patch = raw_patch.squeeze(0)

        if label_patch.ndim == 4:
            label_patch = label_patch.unsqueeze(0)
            remove_dims = 1
        else:
            label_patch = label_patch.unsqueeze(0).unsqueeze(0)
            remove_dims = 2

        label_dtype = label_patch.dtype
        # check if label patch is of torch int type
        if label_dtype in [torch.int, torch.int8, torch.int16, torch.int32, torch.int64]:
            # convert to float for interpolation
            label_patch = label_patch.float()

        # interpolate label patch
        label_patch = interpolate(label_patch, self.patch_shape, mode="nearest")

        # remove additional dimensions
        for _ in range(remove_dims):
            label_patch = label_patch.squeeze(0)

        # convert back to int if necessary
        if label_dtype in [torch.int, torch.int8, torch.int16, torch.int32, torch.int64]:
            if label_dtype == torch.int64:
                label_patch = label_patch.long()
            else:
                label_patch = label_patch.int()

        return raw_patch, label_patch

    def _apply_offsets(self, idx: tuple, offsets: list, is_start: bool) -> tuple:
        if len(idx) == 4:
            spatial_idx = idx[1:]
        else:
            spatial_idx = idx

        new_idx = []
        for i, o, s in zip(spatial_idx, offsets, self.volume_shape, strict=True):
            if is_start:
                # prevent negative start
                start = max(0, i.start + o)
                stop = i.stop
            else:
                start = i.start
                # prevent stop exceeding the volume shape
                stop = min(s, i.stop + o)

            new_idx.append(slice(start, stop))

        if len(idx) == 4:
            return (idx[0],) + tuple(new_idx)

        return tuple(new_idx)


class ConfigDataset(Dataset):
    """
    Abstract class for datasets that are configured via a dictionary.
    """

    def __getitem__(self, index) -> np.ndarray:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    @classmethod
    def create_datasets(cls, dataset_config: dict, phase: str) -> list[Dataset]:
        """
        Factory method for creating a list of datasets based on the provided config.

        Args:
            dataset_config: dataset configuration
            phase: one of ['train', 'val', 'test']

        Returns:
            list of `Dataset` instances
        """
        raise NotImplementedError

    @classmethod
    def prediction_collate(cls, batch: list) -> Any:
        """Default collate_fn. Override in child class for non-standard datasets.

        Args:
            batch: List of samples from the dataset.

        Returns:
            Collated batch.
        """
        return default_prediction_collate(batch)


class SliceBuilder:
    """
    Builds the position of the patches in a given raw/label ndarray based on the patch and stride shape.

    Args:
        raw_dataset: raw data
        label_dataset: ground truth labels
        patch_shape: the shape of the patch DxHxW
        stride_shape: the shape of the stride DxHxW
        kwargs: additional metadata
    """

    def __init__(
        self,
        raw_dataset: h5py.Dataset,
        label_dataset: h5py.Dataset,
        patch_shape: tuple[int, int, int],
        stride_shape: tuple[int, int, int],
        **kwargs,
    ):
        patch_shape = tuple(patch_shape)
        stride_shape = tuple(stride_shape)
        skip_shape_check = kwargs.get("skip_shape_check", False)
        if not skip_shape_check:
            self._check_patch_shape(patch_shape)

        self._raw_slices = self._build_slices(raw_dataset, patch_shape, stride_shape)
        if label_dataset is None:
            self._label_slices = None
        else:
            if raw_dataset.ndim != label_dataset.ndim:
                self._label_slices = self._build_slices(label_dataset, patch_shape, stride_shape)
                assert len(self._raw_slices) == len(self._label_slices)
            else:
                # if raw and label have the same dim, they have the same shape and thus the same slices
                self._label_slices = self._raw_slices

    @property
    def raw_slices(self):
        return self._raw_slices

    @property
    def label_slices(self):
        return self._label_slices

    @staticmethod
    def _build_slices(
        dataset: h5py.Dataset, patch_shape: tuple[int, int, int], stride_shape: tuple[int, int, int]
    ) -> list[tuple[slice, ...]]:
        """Iterates over a given n-dim dataset patch-by-patch with a given stride and builds an array of slice positions.

        Args:
            dataset: The dataset to build slices for.
            patch_shape: Shape of the patch.
            stride_shape: Shape of the stride.

        Returns:
            List of slices, i.e. [(slice, slice, slice, slice), ...] if len(shape) == 4
            or [(slice, slice, slice), ...] if len(shape) == 3.
        """
        slices = []
        if dataset.ndim == 4:
            in_channels, i_z, i_y, i_x = dataset.shape
        else:
            i_z, i_y, i_x = dataset.shape

        k_z, k_y, k_x = patch_shape
        s_z, s_y, s_x = stride_shape
        z_steps = SliceBuilder._gen_indices(i_z, k_z, s_z)
        for z in z_steps:
            y_steps = SliceBuilder._gen_indices(i_y, k_y, s_y)
            for y in y_steps:
                x_steps = SliceBuilder._gen_indices(i_x, k_x, s_x)
                for x in x_steps:
                    slice_idx = (
                        slice(z, z + k_z),
                        slice(y, y + k_y),
                        slice(x, x + k_x),
                    )
                    if dataset.ndim == 4:
                        slice_idx = (slice(0, in_channels),) + slice_idx
                    slices.append(slice_idx)
        return slices

    @staticmethod
    def _gen_indices(i, k, s):
        assert i >= k, "Sample size has to be bigger than the patch size"
        for j in range(0, i - k + 1, s):
            yield j
        if j + k < i:
            yield i - k

    @staticmethod
    def _check_patch_shape(patch_shape):
        assert len(patch_shape) == 3, "patch_shape must be a 3D tuple"
        assert patch_shape[1] >= 64 and patch_shape[2] >= 64, "Height and Width must be greater or equal 64"


class FilterSliceBuilder(SliceBuilder):
    """
    Filter patches containing less than the `threshold` of non-zero values.

    Args:
        raw_dataset: raw data
        label_dataset: ground truth labels
        patch_shape: the shape of the patch DxHxW
        stride_shape: the shape of the stride DxHxW
        ignore_index: ignore index in the label dataset; this label will be matched to 0 before filtering
        threshold: the threshold of non-zero values in the label patch
        slack_acceptance: the probability of accepting a patch that does not meet the threshold criteria
        kwargs: additional metadata
    """

    def __init__(
        self,
        raw_dataset: h5py.Dataset,
        label_dataset: h5py.Dataset,
        patch_shape: tuple[int, int, int],
        stride_shape: tuple[int, int, int],
        ignore_index: int | None = None,
        threshold: float = 0.6,
        slack_acceptance: float = 0.01,
        lazy_loader: bool = False,
        **kwargs,
    ):
        super().__init__(raw_dataset, label_dataset, patch_shape, stride_shape, **kwargs)
        if label_dataset is None:
            return
        assert 0 <= threshold <= 1, "Threshold must be in the range [0, 1]"
        assert 0 <= slack_acceptance <= 1, "Slack acceptance must be in the range [0, 1]"

        if not lazy_loader:
            # if not lazy, load label dataset into memory for faster patch filtering
            label_dataset = label_dataset[()]

        rand_state = np.random.RandomState(47)

        def ignore_predicate(raw_label_idx: tuple[slice, slice]) -> bool:
            label_idx = raw_label_idx[1]
            patch = label_dataset[label_idx]
            if ignore_index is not None:
                patch = np.copy(patch)
                patch[patch == ignore_index] = 0
            non_ignore_counts = np.count_nonzero(patch != 0)
            non_ignore_counts = non_ignore_counts / patch.size
            return non_ignore_counts > threshold or rand_state.rand() < slack_acceptance

        zipped_slices = zip(self.raw_slices, self.label_slices, strict=True)
        # ignore slices containing too much ignore_index
        filtered_slices = list(filter(ignore_predicate, zipped_slices))
        # log number of filtered patches
        logger.info(
            f"FilterSliceBuilder: Loading {len(filtered_slices)} out of {len(self.raw_slices)} patches: "
            f"{int(100 * len(filtered_slices) / len(self.raw_slices))}%"
        )
        # unzip and save slices
        raw_slices, label_slices = zip(*filtered_slices, strict=True)
        self._raw_slices = list(raw_slices)
        self._label_slices = list(label_slices)


def _loader_classes(class_name):
    modules = ["pytorch3dunet.datasets.hdf5", "pytorch3dunet.datasets.dsb", "pytorch3dunet.datasets.utils"]
    return get_class(class_name, modules)


def get_slice_builder(raw: h5py.Dataset, label: h5py.Dataset, config: dict) -> SliceBuilder:
    assert "name" in config
    logger.info(f"Slice builder config: {config}")
    slice_builder_cls = _loader_classes(config["name"])
    return slice_builder_cls(raw, label, **config)


def get_train_loaders(config: dict) -> dict[str, DataLoader]:
    """
    Returns dictionary containing the training and validation loaders (torch.utils.data.DataLoader).
    Args:
        config:  a top level configuration object containing the 'loaders' key
    Returns:
        dict {
            'train': <train_loader>
            'val': <val_loader>
        }
    """
    assert "loaders" in config, "Could not find data loaders configuration"
    loaders_config = config["loaders"]
    assert set(loaders_config["train"]["file_paths"]).isdisjoint(loaders_config["val"]["file_paths"]), (
        "Train and validation 'file_paths' overlap. One cannot use validation data for training!"
    )

    logger.info("Creating training and validation set loaders...")

    # get dataset class
    dataset_cls_str = loaders_config.get("dataset", None)
    if dataset_cls_str is None:
        dataset_cls_str = "StandardHDF5Dataset"
        logger.warning(f"Cannot find dataset class in the config. Using default '{dataset_cls_str}'.")
    dataset_class = _loader_classes(dataset_cls_str)

    train_datasets = dataset_class.create_datasets(loaders_config, phase="train")
    val_datasets = dataset_class.create_datasets(loaders_config, phase="val")

    num_workers = loaders_config.get("num_workers", 1)
    logger.info(f"Number of workers for train/val dataloader: {num_workers}")
    batch_size = loaders_config.get("batch_size", 1)
    device = config.get("device", None)
    assert device, "Device not specified in the config file and could not be inferred automatically"
    if device == TorchDevice.CUDA and torch.cuda.device_count() > 1:
        logger.info(
            f"{torch.cuda.device_count()} GPUs available. Using batch_size = {torch.cuda.device_count()} * {batch_size}"
        )
        batch_size = batch_size * torch.cuda.device_count()

    logger.info(f"Batch size for train/val loader: {batch_size}")
    loader_kwargs = os_dependent_dataloader_kwargs()
    return {
        "train": DataLoader(
            ConcatDataset(train_datasets),
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            **loader_kwargs,
        ),
        # don't shuffle during validation: useful when showing how predictions for a given batch get better over time
        "val": DataLoader(
            ConcatDataset(val_datasets),
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            **loader_kwargs,
        ),
    }


def get_test_loaders(config: dict) -> DataLoader:
    """Returns test DataLoader.

    Args:
        config: A top level configuration object containing the 'loaders' key.

    Returns:
        Generator of DataLoader objects.
    """

    assert "loaders" in config, "Could not find data loaders configuration"
    loaders_config = config["loaders"]

    logger.info("Creating test set loaders...")

    # get dataset class
    dataset_cls_str = loaders_config.get("dataset", None)
    if dataset_cls_str is None:
        dataset_cls_str = "StandardHDF5Dataset"
        logger.warning(f"Cannot find dataset class in the config. Using default '{dataset_cls_str}'.")
    dataset_class = _loader_classes(dataset_cls_str)

    test_datasets = dataset_class.create_datasets(loaders_config, phase="test")

    num_workers = loaders_config.get("num_workers", 1)
    logger.info(f"Number of workers for the dataloader: {num_workers}")

    batch_size = loaders_config.get("batch_size", 1)
    device = config.get("device", None)
    assert device, "Device not specified in the config file and could not be inferred automatically"
    if device == TorchDevice.CUDA and torch.cuda.device_count() > 1:
        logger.info(
            f"{torch.cuda.device_count()} GPUs available. Using batch_size = {torch.cuda.device_count()} * {batch_size}"
        )
        batch_size = batch_size * torch.cuda.device_count()

    logger.info(f"Batch size for dataloader: {batch_size}")

    # use generator in order to create data loaders lazily one by one
    for test_dataset in test_datasets:
        if hasattr(test_dataset, "prediction_collate"):
            collate_fn = test_dataset.prediction_collate
        else:
            collate_fn = default_prediction_collate

        dataloader_kwargs = os_dependent_dataloader_kwargs()

        yield DataLoader(
            test_dataset, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn, **dataloader_kwargs
        )


def default_prediction_collate(batch: list) -> Any:
    """Default collate_fn to form a mini-batch of Tensor(s) for HDF5 based datasets.

    Args:
        batch: List of samples from the dataset.

    Returns:
        Collated batch.
    """
    error_msg = "batch must contain tensors or slice; found {}"
    if isinstance(batch[0], torch.Tensor):
        return torch.stack(batch, 0)
    elif isinstance(batch[0], tuple) and isinstance(batch[0][0], slice):
        return batch
    elif isinstance(batch[0], collections.abc.Sequence):
        transposed = zip(*batch, strict=True)
        return [default_prediction_collate(samples) for samples in transposed]

    raise TypeError(error_msg.format(type(batch[0])))


def calculate_stats(img: np.array, skip: bool = False) -> dict[str, Any]:
    """
    Calculates the minimum percentile, maximum percentile, mean, and standard deviation of the image.

    Args:
        img: The input image array.
        skip: if True, skip the calculation and return None for all values.

    Returns:
        tuple[float, float, float, float]: The minimum percentile, maximum percentile, mean, and std dev
    """
    if not skip:
        pmin, pmax, mean, std = np.percentile(img, 1), np.percentile(img, 99.6), np.mean(img), np.std(img)
    else:
        pmin, pmax, mean, std = None, None, None, None

    return {"pmin": pmin, "pmax": pmax, "mean": mean, "std": std}


def mirror_pad(image: np.ndarray, padding_shape: tuple[int, int, int]) -> np.ndarray:
    """
    Pad the image with a mirror reflection of itself.

    This function is used on data in its original shape before it is split into patches.

    Args:
        image (np.ndarray): The input image array to be padded.
        padding_shape (tuple of int): Specifies the amount of padding for each dimension, should be YX or ZYX.

    Returns:
        np.ndarray: The mirror-padded image.

    Raises:
        ValueError: If any element of padding_shape is negative.
    """
    assert len(padding_shape) == 3, "Padding shape must be specified for each dimension: ZYX"

    if any(p < 0 for p in padding_shape):
        raise ValueError("padding_shape must be non-negative")

    if all(p == 0 for p in padding_shape):
        return image

    pad_width = [(p, p) for p in padding_shape]

    if image.ndim == 4:
        pad_width = [(0, 0)] + pad_width
    return np.pad(image, pad_width, mode="reflect")


def remove_padding(m: np.ndarray, padding_shape: int | tuple[int, ...] | None) -> np.ndarray:
    """
    Removes padding from the margins of a multi-dimensional array.

    Args:
        m (np.ndarray): The input array to be unpadded.
        padding_shape (tuple of int, optional): The amount of padding to remove from each dimension.
            Assumes the tuple length matches the array dimensions.

    Returns:
        np.ndarray: The unpadded array.
    """
    if padding_shape is None:
        return m

    # Correctly construct slice objects for each dimension in padding_shape and apply them to m.
    return m[(..., *(slice(p, -p or None) for p in padding_shape))]
